#include "Mesh.h"

bool Mesh::loadPnt(char* meshFile)
{
	cout << "\n\nMesh initializing (to " << meshFile << ")...\n";

	FILE* fPtr;
	if (! (fPtr = fopen(meshFile, "r")))
	{
		cout << "cannot read " << meshFile << endl;
		return false;
		//exit(0);
	}


	while (! feof( fPtr ) )
	{
		float a, b, c;
		fscanf(fPtr, "%f %f %f\n", &a, &b, &c);
		float* co = new float[3];
		co[0] = a;
		co[1] = b;
		co[2] = c;
		addVertexND(co); //ND: no duplicate check
	}

	cout << "Mesh has " << (int) tris.size() << " tris, " << (int) verts.size() << " verts, " << (int) edges.size() << " edges\nInitialization done\n";
	return true;
}

int Mesh::addVertex(float x, float y, float z)
{
	float v[3] = {x, y, z};
	return addVertex(v);
}
int Mesh::addVertex(float* v)
{
	//add v to verts, if not duplicated

	bool duplicateExists = false;
	int i = (int) verts.size();
	
	//duplicate prevention
	float ROUNDING_ERR = 0.01f;
	for (i = (int) verts.size() - 1; i >= 0; i--)
	{
		if (fabs(verts[i]->coords[0] - v[0]) < ROUNDING_ERR && //since i scale stuff initially w/ SCALE_FACTOR = 1000 just in case, ROUNDING_ERR = 0.01f should be good here
			fabs(verts[i]->coords[1] - v[1]) < ROUNDING_ERR &&
			fabs(verts[i]->coords[2] - v[2]) < ROUNDING_ERR)
		{
			duplicateExists = true;
			break;	//more than efficiency: return the correct i by this break
		}
	}
	if (! duplicateExists)
	{
		int idx = (int) verts.size();
		verts.push_back(new Vertex(idx, v));
		return idx; //pointer to this new item (last index)
	}

	//this duplicate vert v is ith element of verts
	return i;
}

void Mesh::addVertexND(float* c)
{
	//fastly add verts w/ coords c to mesh; ND: no duplication check

	int vSize = (int) verts.size(); //size before push_back just-below
	verts.push_back(new Vertex(vSize, c));
}

int Mesh::addTriangle(int v1i, int v2i, int v3i)
{
	//add tri v1i-v2i-v3i to the triInd'th element of tris

	int triInd = (int) tris.size();

	tris.push_back( new Triangle(triInd, v1i, v2i, v3i) );

	//update each vertex with its new tri and vert neighbors
	verts[v1i]->triNeighbors.push_back(triInd);
	//addVertNeighbor returns true if verts[v1i] is not neighbor of [v2i]; so they can form a new edge
	if (verts[v1i]->addVertNeighbor(v2i))	addEdge(v1i, v2i);
	if (verts[v1i]->addVertNeighbor(v3i))	addEdge(v1i, v3i);
	//in other words, a 'false' returned by addVertNeighbor prevents adding the same edge if it already exists

	verts[v2i]->triNeighbors.push_back(triInd);
	verts[v2i]->addVertNeighbor(v1i);	//v1i v2i edge is (probably) added above; do not add v2i v1i edge!!!
	if (verts[v2i]->addVertNeighbor(v3i))	addEdge(v2i, v3i);	//v3i v2i is not added above, so add this if ok

	verts[v3i]->triNeighbors.push_back(triInd);
	verts[v3i]->addVertNeighbor(v1i);	//no addEdge risk, so no if control
	verts[v3i]->addVertNeighbor(v2i);

	return triInd;
}

int Mesh::getEdgeSharedBy(int vi, int vj)
{
	//return idx of edge[vi, vj]

	for (unsigned int vie = 0; vie < verts[vi]->edgeList.size(); vie++)
		for (unsigned int vje = 0; vje < verts[vj]->edgeList.size(); vje++)
			if (verts[vj]->edgeList[vje] == verts[vi]->edgeList[vie])
				return verts[vj]->edgeList[vje];

	//edge not exist
	return -99;
}

inline float distanceBetween(float* v1, float* v2)
{
	//Euclidean distance b/w v1 & v2
	return (float) sqrt( (v2[0]-v1[0])*(v2[0]-v1[0]) + (v2[1]-v1[1])*(v2[1]-v1[1]) + (v2[2]-v1[2])*(v2[2]-v1[2]));
}

void Mesh::addEdge(int v1i, int v2i, bool interiorEdge)
{
	int initial = (int) edges.size();
	Edge* newcomer = new Edge(initial, v1i, v2i, distanceBetween(verts[v1i]->coords, verts[v2i]->coords), interiorEdge);

	//update the list of edges of which verts[v1i] is a member
	//for geodesic computations in dijkstraShortestPaths() i will discard interior edges (v.interiorEdgeList) as geodesic is defined as surface distance and interior edges are just noises distracting the true geodesics; interiorEdges however are necessary for volumetric regulation term in spring system and deformations
	if (interiorEdge)
	{
		verts[v1i]->interiorEdgeList.push_back(newcomer->idx);
		verts[v2i]->interiorEdgeList.push_back(newcomer->idx); //do same thing for [v2i]
	}
	else
	{
		verts[v1i]->edgeList.push_back(newcomer->idx);
		verts[v2i]->edgeList.push_back(newcomer->idx); //do same thing for [v2i]
	}

	edges.push_back(newcomer);

	//information
	if (newcomer->length < minEdgeLen)
		minEdgeLen = newcomer->length;
	if (newcomer->length > maxEdgeLen)
	{
		maxEdgeLen = newcomer->length;
//cout << v1i << " " << v2i << "\t" << newcomer->length << endl;
	}
	edgeLenTotal += newcomer->length;
}

int* Mesh::getTrianglesSharedBy(Edge* bigEdge, int v1i, int v2i)
{
	//returns 2 tris adjacent to bigEdge (where endpnts of bigEdge is v1i and v2i)

	//tri t is shared by bigEdge iff it exists in the triNeigbors of verts[v1i] & [v2i]
	if (v1i == -1) //learn v1i and v2i from bigEdge
	{
		v1i = bigEdge->v1i;
		v2i = bigEdge->v2i;
	}
	int a = 0, tr1, tr2;
	int* shTris = new int[2];	//portable version (memo-friendly as well 'cos i delete shrdTris in split/flip())
	//seeing this took my 6 hours!!! prevent removed triangles from being selected as shared!
	for (unsigned int t1 = 0; t1 < verts[v1i]->triNeighbors.size(); t1++)
	{
		tr1 = verts[v1i]->triNeighbors[t1];
		for (unsigned int t2 = 0; t2 < verts[v2i]->triNeighbors.size(); t2++)
		{
			tr2 = verts[v2i]->triNeighbors[t2];
			if  (tris[tr2] != 0 && (tr1 == tr2))
				shTris[a++] = tr1; //or tr2
		}
	}
	//if called from split/flip, a is certainly incremented by 1; but it remains at 1 (no more ++) 
	//if bigEdge is a border edge (not in my case 'cos my mesh is closed); take precautions anyway;
	if (a == 1) shTris[1] = -1;	//number -1 must not change 'cos used by somewhere else
if (shTris[1] == -1) cout << v1i << "," << v2i << " is a border edge found in my closed-mesh :(\n\n"; //exit(0);}

//if called from contract, a may remain 0, which tells me to remove that edgesAdded[cea] neighbor to 2 removed tris
if (a == 0)	shTris[0] = -1;	//since that part of contract() is inactive, don't care here

//assert(shTris[0] != shTris[1]);
	return shTris;
}

/*void Mesh::addTetrahedron(int v1i, int v2i, int v3i, int v4i)
{
	//add tetrahedron v1i-v2i-v3i-v4i, along with the tris they imply, to the tetInd'th element of tets and add this tet to the tetNeighbors of the forming vertices

	//implied tris
	int t1i = addTriangleTetmesh(v2i, v3i, v4i, true), //according to right-handed orientation of .ele files (http://tetgen.berlios.de/fformats.ele.html)
		t2i = addTriangleTetmesh(v1i, v2i, v4i, true),
		t3i = addTriangleTetmesh(v1i, v3i, v2i, true),
		t4i = addTriangleTetmesh(v1i, v4i, v3i, true),
		tetInd = (int) tets.size();
	tets.push_back( new Tetrahedron(tetInd, t1i, t2i, t3i, t4i, v1i, v2i, v3i, v4i) );
	verts[v1i]->tetNeighbors.push_back(tetInd);
	verts[v2i]->tetNeighbors.push_back(tetInd);
	verts[v3i]->tetNeighbors.push_back(tetInd);
	verts[v4i]->tetNeighbors.push_back(tetInd);
}*/

float Mesh::ICP(Mesh* dtMesh, bool adaptiveScaling, int nMaxIters)
{
	//transform this mesh towards mesh2 using ICP; closed-form rotation matrix is from eq. 21 of the original paper: A Method for Registration of 3-D Shapes

	cout << "\n\nICP in action..\n";

	//center of masses of the k-d point sets
	int k = 3, nVerts = (int) verts.size(), nVoxels = (int) dtMesh->verts.size();
//assert(nVoxels >= nVerts); //small-size mesh is moving towards fixed voxels by establishing nVerts-size closest pnt correspondences w/ which the crucial cross-covariance mtrx is filled (some voxels may remain unmatched and corresp may be many to one)
	float* m1 = new float[3], * m2 = new float[3];
	float sum[3] = {0.0f, 0.0f, 0.0f};
	for (int i = 0; i < nVerts; i++)
		for (int c = 0; c < 3; c++)
			sum[c] += verts[i]->coords[c];
	for (int c = 0; c < 3; c++)
		m1[c] = sum[c] / (float) nVerts;
	sum[0] = sum[1] = sum[2] = 0.0f;
	for (int i = 0; i < nVoxels; i++)
		for (int c = 0; c < 3; c++)
			sum[c] += dtMesh->verts[i]->coords[c];
	for (int c = 0; c < 3; c++)
		m2[c] = sum[c] / (float) nVoxels;
//	cout << "center of masses of mesh & voxels = " << m1[0] << ", " << m1[1] << ", " << m1[2] << "  &  " << m2[0] << ", " << m2[1] << ", " << m2[2] << "\n\n"; 

	MatrixXf ms1(k, nVerts), ms2(k, nVerts); //mean-shifted kxnVerts matrix for the mesh and matched voxels; X is for dynamic size f is for float
	Matrix3f cov12; //kxk cross-covariance mtrx of 2 sets, mesh & voxels where k = 3
	Matrix4f Q; //complicated mtrx created using cov12

	float** R1 = new float*[3]; //desired 3x3 rotation mtrx of current iteration
	for (int i = 0; i < 3; i++)
		R1[i] = new float[3];
	float* t1 = new float[3]; //desired 1x3 translation
	float** R = new float*[3], //accumulated 3x3 rotation mtrx of all iterations
		 ** Rtmp = new float*[3]; //auxiliary for R
	for (int i = 0; i < 3; i++)
	{
		R[i] = new float[3];
		Rtmp[i] = new float[3];
		for (int j = 0; j < 3; j++)
			R[i][j] = (i == j ? 1.0f : 0.0f);
	}

#ifdef KDTREE_FOR_CLOSEST_PNTS
	//k-d tree holding the voxel coordinates and their corresponding idxs to speed up the closest pnt search below
	kdtree* kd = kd_create(3);
	kdres* resultSet = NULL;
	int* voxelIdx;
	float voxelCoord[3];
	for (int i = 0; i < nVoxels; i++)
	{
		int* tmp = new int[1];
		*tmp = i;
		kd_insert3f(kd, dtMesh->verts[i]->coords[0], dtMesh->verts[i]->coords[1], dtMesh->verts[i]->coords[2], tmp);
	}
#endif

	//parameters of the linear system to be solved in adaptiveScaling mode
	float A, B, * C = new float[3], * D = new float[3];	
	//iteration parameters
	int nIters = 0;
	float mse = INF, prevMse = 0.0f, //mean-square error objective function to be minimized
		  alignErr = 100.0f, closeness = 0.0001f; //mse==prevMse test w/ closeness threshold; early-converge if mse value not change from the previous iteration
	while (++nIters <= nMaxIters)
	{
		if (//nIters > 70 && //activate closeness test after a good deal of iters; a heuristic no in use
			fabs(mse - prevMse) < closeness)
			break;

		//new center of mass of the transforming mesh
		sum[0] = sum[1] = sum[2] = 0.0f;
		for (int v = 0; v < nVerts; v++)
			for (int c = 0; c < 3; c++)
				sum[c] += verts[v]->coords[c];
		for (int c = 0; c < 3; c++)
			m1[c] = sum[c] / (float) nVerts;

		//before filling ms1, i need closest-pnt correspondence b/w mesh & voxels
		float dist; //closest w.r.t. Euclidean distance
		vector< int > matchedVoxels; //idxs of voxels that are matched to 1+ mesh verts
#ifdef KDTREE_FOR_CLOSEST_PNTS
		//using a k-d tree for nVerts x log(nVoxels) performance instead of nVerts x nVoxels of the simple exhaustive search below
		for (int bv1 = 0; bv1 < nVerts; bv1++) //for each mesh verts
		{
			resultSet = kd_nearest3f(kd, verts[bv1]->coords[0], verts[bv1]->coords[1], verts[bv1]->coords[2]); //find the voxel closest to verts[bv1]->coords
			voxelIdx = (int*) kd_res_itemf(resultSet, voxelCoord); //get/retrieve the data and position of the current result item; not interested in the voxelCoord though
			dtMesh->verts[*voxelIdx]->closestVertIdx = bv1;
			matchedVoxels.push_back(*voxelIdx);
		}
#else
		//exhaustive search for closest matches in quadratic time
		for (int bv1 = 0; bv1 < nVerts; bv1++) //for each mesh verts
		{
			float minDist = INF;
			int mv = -1;
			for (int bv2 = 0; bv2 < nVoxels; bv2++) //find its closest voxel match
			{
				dist = distance2Between(dtMesh->verts[bv2]->coords, verts[bv1]->coords); //omit sqrt() to save time
				if (dist < minDist)
				{
					minDist = dist;
					mv = bv2;
				}
			}
			matchedVoxels.push_back(mv);
			dtMesh->verts[mv]->closestVertIdx = bv1;
		}
#endif
assert((int) matchedVoxels.size() == nVerts); //matchedVoxels[i] is the voxel matched to mesh verts[i]
		//now that closest pnt correspondence done, i.e. closestVertIdx's are set, i can fill ms1
		//fill mean-shifted kxnVerts ms1 matrix for the 1st set (mesh)
		for (int i = 0; i < k; i++) //for all k rows
			for (int j = 0; j < nVerts; j++) //fill columns
				ms1(i, j) = verts[ dtMesh->verts[ matchedVoxels[j] ]->closestVertIdx ]->coords[i] - m1[i];
		//fill mean-shifted 3xnVerts matrix for the 2nd set (matched voxels, hence of size nVerts for sure)
		//new center of mass of the newly matched voxels
		sum[0] = sum[1] = sum[2] = 0.0f;
		for (int v = 0; v < nVerts; v++)
			for (int c = 0; c < 3; c++)
				sum[c] += dtMesh->verts[ matchedVoxels[v] ]->coords[c];
		for (int c = 0; c < 3; c++)
			m2[c] = sum[c] / (float) nVerts;
		//fill mean-shifted kxnVerts ms2 matrix for the voxels; X is for dynamic size f is for float
		for (int i = 0; i < k; i++) //for all k rows
			for (int j = 0; j < nVerts; j++) //fill n1 columns
				ms2(i, j) = dtMesh->verts[ matchedVoxels[j] ]->coords[i] - m2[i];
		MatrixXf ms2t = ms2.transpose(); //transpose of ms

		//kxk cross-covariance mtrx b/w mesh and voxels
		cov12 = ms1 * ms2t; //k x nVerts * nVerts x k
		for (int i = 0; i < 3; i++)
			for (int j = 0; j < 3; j++)
				cov12(i, j) /= nVerts;

		//Q = 4x4 complicated mtrx, see the paper
		Q(0, 0) = cov12(0, 0) + cov12(1, 1) + cov12(2, 2);
		Q(1, 0) = Q(0, 1) = cov12(1, 2) - cov12(2, 1); //(paper notation: S23 - S32, say S to cov12, and idxs of paper, i.e. start from 1)
		Q(2, 0) = Q(0, 2) = cov12(2, 0) - cov12(0, 2); //S31 - S13
		Q(3, 0) = Q(0, 3) = cov12(0, 1) - cov12(1, 0); //S12 - S21
		Q(1, 1) = 2*cov12(0, 0) - Q(0, 0);		Q(1, 2) = cov12(0, 1) + cov12(1, 0);	Q(1, 3) = cov12(0, 2) + cov12(2, 0);
		Q(2, 1) = cov12(1, 0) + cov12(0, 1);	Q(2, 2) = 2*cov12(1, 1) - Q(0, 0);		Q(2, 3) = cov12(1, 2) + cov12(2, 1);
		Q(3, 1) = cov12(2, 0) + cov12(0, 2);	Q(3, 2) = cov12(2, 1) + cov12(1, 2);	Q(3, 3) = 2*cov12(2, 2) - Q(0, 0);
		//desired rotation is decided by 1x4 quaternion vector, which is the (largest eigval's) eigvec of Q
		SelfAdjointEigenSolver<Matrix4f> eigensolver(Q); //solve eigenvalue/vector for symmetric n by n matrices, a.k.a. selfadjoint matrices
		if (eigensolver.info() != Success) abort();
		Vector4f eig_vals = eigensolver.eigenvalues(); //given in ascending order
		Matrix4f eig_vecs = eigensolver.eigenvectors(); //last column of eig_vecs corresponds to the largest eigenvalue, hence the one i want
		
		//1x4 quaternion = [q0, q1, q2, q3]
		float q0 = eig_vecs(0, 3), q1 = eig_vecs(1, 3), q2 = eig_vecs(2, 3), q3 = eig_vecs(3, 3);
		//if (q0 < 0) { q0 = -q0;	q1 = -q1;	q2 = -q2;	q3 = -q3; } //needed??????????

		//fill desired 3x3 rotation mtrx (i'm used to 4x4 mtrxs w/ coords but this 3x3 works 'cos of quaternions, i guess)
		R1[0][0] = q0*q0 + q1*q1 - q2*q2 - q3*q3;	R1[0][1] = 2*(q1*q2 - q0*q3);				R1[0][2] = 2*(q1*q3 + q0*q2);
		R1[1][0] = 2*(q1*q2 + q0*q3);				R1[1][1] = q0*q0 + q2*q2 - q1*q1 - q3*q3;	R1[1][2] = 2*(q2*q3 - q0*q1);
		R1[2][0] = 2*(q1*q3 - q0*q2);				R1[2][1] = 2*(q2*q3 + q0*q1);				R1[2][2] = q0*q0 + q3*q3 - q1*q1 - q2*q2;
		float uniformScaler = 1.0f;
		if (! adaptiveScaling) //classical ICP that doesn't bother w/ uniform scaling
		{
			//fill desired 1x3 translation
			t1[0] = m2[0] - (R1[0][0]*m1[0] + R1[0][1]*m1[1] + R1[0][2]*m1[2]);
			t1[1] = m2[1] - (R1[1][0]*m1[0] + R1[1][1]*m1[1] + R1[1][2]*m1[2]);
			t1[2] = m2[2] - (R1[2][0]*m1[0] + R1[2][1]*m1[1] + R1[2][2]*m1[2]);

			//apply transformation of this iteration to all verts of the transforming mesh
			for (int v = 0; v < nVerts; v++)
			{
				float x = verts[v]->coords[0],
					  y = verts[v]->coords[1],
					  z = verts[v]->coords[2];
				//rotation
				verts[v]->coords[0] = R1[0][0]*x + R1[0][1]*y + R1[0][2]*z;
				verts[v]->coords[1] = R1[1][0]*x + R1[1][1]*y + R1[1][2]*z;
				verts[v]->coords[2] = R1[2][0]*x + R1[2][1]*y + R1[2][2]*z;
				//translation
				verts[v]->coords[0] += t1[0];
				verts[v]->coords[1] += t1[1];
				verts[v]->coords[2] += t1[2];
			}			
		}
		else //our ICP that solves for translation and uniform scaling simulataneously to improve the initial scale guess as well
		{
			//the math of this part is explained in my paper titled Skuller; check it out
			A = B = 0.0f;
			for (int i = 0; i < 3; i++)
				C[i] = D[i] = 0.0f;
			//apply just ICP rotation of this iteration to all verts of the transforming mesh
			for (int v = 0; v < nVerts; v++)
			{
				float x = verts[v]->coords[0],
					  y = verts[v]->coords[1],
					  z = verts[v]->coords[2];
				//rotation
				verts[v]->coords[0] = R1[0][0]*x + R1[0][1]*y + R1[0][2]*z;
				verts[v]->coords[1] = R1[1][0]*x + R1[1][1]*y + R1[1][2]*z;
				verts[v]->coords[2] = R1[2][0]*x + R1[2][1]*y + R1[2][2]*z;

				//D = [Dx Dy Dz] is computed Dx is the sum of x-coords of all matched voxel points, and so on (since matched voxels change each iteration D is not fixed); similarly C = [Cx Cy Cz] for transformed, i.e. ICP-rotated, mesh points
				for (int i = 0; i < 3; i++)
				{
					D[i] += dtMesh->verts[ matchedVoxels[v] ]->coords[i];
					C[i] += verts[v]->coords[i];
				}
				A += verts[v]->coords[0]*verts[v]->coords[0] + verts[v]->coords[1]*verts[v]->coords[1] + verts[v]->coords[2]*verts[v]->coords[2];
				B += verts[v]->coords[0]*dtMesh->verts[ matchedVoxels[v] ]->coords[0] + verts[v]->coords[1]*dtMesh->verts[ matchedVoxels[v] ]->coords[1] + verts[v]->coords[2]*dtMesh->verts[ matchedVoxels[v] ]->coords[2];
			}
			//solve for Px = r linear system to decide x = [s t1.x t1.y t1.z] where s is the uniform scale factor, t1[] is the rotation
			Matrix4f P;
			P << A,C[0],C[1],C[2],	C[0],(float)nVerts,0.0f,0.0f,	C[1],0.0f,(float)nVerts,0.0f,	C[2],0.0f,0.0f,(float)nVerts;
//cout << "Here is the matrix P:\n" << P << endl;
			Vector4f r;
			r << B,D[0],D[1],D[2];
//cout << "Here is the right hand side r:\n" << r << endl;
			Vector4f x = P.jacobiSvd(ComputeFullU | ComputeFullV).solve(r); //s = x(0) is the uniform scale factor, and the rest is the translation t1
//cout << "The least-squares solution is:\n" << x << "\nsame as r?\n" << P * x << endl;
			//apply translation and scaling
			for (int v = 0; v < nVerts; v++)
			{
				//scaling
				verts[v]->coords[0] *= x(0);
				verts[v]->coords[1] *= x(0);
				verts[v]->coords[2] *= x(0);

				//translation
				verts[v]->coords[0] += x(1);
				verts[v]->coords[1] += x(2);
				verts[v]->coords[2] += x(3);				
			}
			//prepare for fprint below
			uniformScaler = x(0);
			for (int i = 0; i < 3; i++)
				t1[i] = x(i+1);
//cout << x(0) << " " << x(1) << " " << x(2) << " " << x(3) << "\n";cout << x(0) << " ";
		}

		prevMse = mse; //mse of previous iteration as a termination condition
		//compute new mse value as an iteration condition
		mse = 0.0f;
		for (int bv2 = 0; bv2 < nVerts; bv2++)
		{
			//squared distance b/w (target) matched voxel and its closest match in mesh whose coords just updated above
			dist = pow(dtMesh->verts[ matchedVoxels[bv2] ]->coords[0] - verts[ dtMesh->verts[ matchedVoxels[bv2] ]->closestVertIdx ]->coords[0], 2.0f) +
				   pow(dtMesh->verts[ matchedVoxels[bv2] ]->coords[1] - verts[ dtMesh->verts[ matchedVoxels[bv2] ]->closestVertIdx ]->coords[1], 2.0f) +
				   pow(dtMesh->verts[ matchedVoxels[bv2] ]->coords[2] - verts[ dtMesh->verts[ matchedVoxels[bv2] ]->closestVertIdx ]->coords[2], 2.0f);
			mse += dist; //no sqrt() above for efficiency, hence dist = squared dist indeed
		}
		mse /= nVerts;
		cout << "mse of ICP iter" << nIters-1 << ": " << mse << "\tmse diff\t" << fabs(mse - prevMse) << endl;
	} //end of while nIters/mse

	//recapture memo
	for (int i = 0; i < 3; i++) delete [] R1[i]; delete [] R1;
	delete [] t1;
	for (int i = 0; i < 3; i++) delete [] R[i]; delete [] R;
	delete [] C;
	delete [] D;
#ifdef KDTREE_FOR_CLOSEST_PNTS
	if (resultSet)
		kd_res_free(resultSet);
	kd_free(kd);
#endif

	cout << nIters-1 << "'th ICP iteration w/ final mse: " << mse << "\nICP done!\n\n";
	return mse;
}

void Mesh::resultToFile(char* fName)
{
	//fprints transformed mesh

	FILE* fPtrS = fopen(fName, "w");
	for (int v = 0; v < (int) verts.size(); v++)
		fprintf(fPtrS, "%f %f %f\n", verts[v]->coords[0], verts[v]->coords[1], verts[v]->coords[2]);
	fclose(fPtrS);

	cout << "\ntransformed mesh fprinted to " << fName << endl;
}
